#!/usr/bin/perl
#############################################################################
#                                                                           #
#  Sync with a Yoma Addonsync repository                                    #
#                                                                           #
#  Version: 0.4.1                                                           #
#  Author:  HELLinG3R <helling3r@silencers.de>                              #
#                                                                           #
#  Required modues are:                                                     #
#    Cwd                                                                    #
#    Getopt::Std                                                            #
#    File::Basename, :Path, ::Copy, ::GLob, ::Fetch, ::Find                 #
#    Digest::MD5                                                            #
#    XML::Simple                                                            #
#  These modules are usually already available at your site.                #
#                                                                           #
#  Required external applications:                                          #
#    7Zip     (as command "7zr")                                            #
#                                                                           #
#                                                                           #
#############################################################################
#  This program is free software; you can redistribute it and/or modify     #
#  it under the terms of the GNU General Public License as published by     #
#  the Free Software Foundation; either version 2 of the License, or        #
#  (at your option) any later version.                                      #
#                                                                           #
#  This program is distributed in the hope that it will be useful,          #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#  GNU General Public License for more details.                             #
#                                                                           #
#  You should have received a copy of the GNU General Public License        #
#  along with this program; if not, write to the Free Software              #
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.   #
#############################################################################

use strict;
use Getopt::Std;
use File::Path;
use File::Basename;
use File::Copy;
use File::Glob;
use File::Fetch;
use File::Find;
use Digest::MD5;
use XML::Simple;
use Cwd qw(getcwd abs_path);

#####################
# Optionsmanagement #
#####################

# Defaults:
my $target_dir      = '.';
my @selected_mods = ('@*'); # array with glob pattern
my $tmp_dir       = '/tmp/';
my $verbose       = 0;
my $debug         = 0;
my $url           = "";
my $sync_mode     = 0;
my $check_mode    = 0;
my $list_mode     = 0;
my $del_orphan    = 0;
my $key_copy      = 0;

# Parse options using Getopt:
my %Options;
my $arg_count = @ARGV;
my $arg_ok = getopts('vVhkSLCda:m:u:', \%Options);
if ($Options{h}) {
        help();
        exit 0;
}
if (!$arg_ok) {
        usage();
        exit 9;
}

# check for missing mandatory parameters
foreach my $o ('u') {
        if ( !defined($Options{$o}) ) {
                print STDERR "Parameter -$o is mandatory!\n";
                usage();
                exit(9);
        }
}

# check for external programs
foreach my $prog ('7zr') {
	`which $prog >/dev/null 2>&1`;
	if ($? > 0) {
		print STDERR "unable to locate '$prog', please make sure it's in your PATH.\n";
		exit(9);
	}
}

if ($Options{d} && $Options{C}) {
	print STDERR "-C and -d are mutually exclusive!\n";
	usage();
	exit(9);
}

if ($Options{d} && $Options{L}) {
        print STDERR "-L and -d are mutually exclusive!\n";
        usage();
        exit(9);
}

if ($Options{k} && $Options{C}) {
        print STDERR "-C and -k are mutually exclusive!\n";
        usage();
        exit(9);
}

if ($Options{k} && $Options{L}) {
        print STDERR "-L and -k are mutually exclusive!\n";
        usage();
        exit(9);
}

# set and verify options
if ($Options{a}){   $target_dir = $Options{a};}
if ($Options{C}){   $check_mode = 1;}
if ($Options{d}){   $del_orphan = 1;}
if ($Options{u}){   $url = $Options{u};}
if ($Options{v}){   $verbose = 1;}
if ($Options{V}){   $debug = 1;}
if ($Options{L}){   $list_mode = 1;}
if ($Options{S}){   $sync_mode = 1;}
if ($Options{k}){   $key_copy = 1;}
if ($Options{m}){
	@selected_mods = split(/;/, $Options{m});
}

# if no mode selection occurred, use default mode
if ($sync_mode + $check_mode + $list_mode == 0) {
	$sync_mode = 1;
}

# if more than one mode is selected, complain
if ($sync_mode + $check_mode + $list_mode > 1) {
	print STDERR "Only one mode can be selected!\n";
        usage();
        exit(9);
}

# set $target_dir to a full system path if it is given relative
if ($target_dir !~ /^\//) {
        $target_dir =~ s/^\.//;
        $target_dir = dirname(abs_path($0))."/$target_dir";
}
# verify existence of target path
if (! -d $target_dir && !$check_mode) {
        if ($verbose) { print STDERR "creating target_dir='$target_dir'\n"; }
        mkpath("$target_dir");
}

# target location of keyfiles
my $keyfolder = "$target_dir/keys/";



##############
# Here we go #
##############

# Try to get the repository metadata 7z.
my $tmp_dir_exp = "$tmp_dir/yas_client-" . "$$/";
if (-d "$tmp_dir_exp") {
        # is existing, clenup
        if ($verbose) { print STDERR "clearing existing tmp_dir_exp='$tmp_dir_exp'\n"; }
        rmtree("$tmp_dir_exp");
}
if ($verbose) { print STDERR "creating tmp_dir_exp='$tmp_dir_exp'\n"; }
mkpath("$tmp_dir_exp");

print STDERR "analyzing repository metadata\n";

print STDERR "  fetching '$url'... ";
my $repo = File::Fetch->new(uri => $url);
my $where = $repo->fetch( to => $tmp_dir_exp );
if (!$where) {
	print STDERR "error!\n";
	rmtree("$tmp_dir_exp");
	exit(1);
} else {
	print STDERR "OK\n";
}

# extract repo name from file
my $repo_name = $repo->file;
$repo_name =~ s/\.7z$//;
if ($verbose) { print STDERR "  extracted repo name='$repo_name'\n";}

# extract the repository metadata
my $owd = getcwd;
if ($verbose) { print STDERR "  chdir to $tmp_dir_exp\n";}
chdir($tmp_dir_exp);
my $zip_cmd = "7zr -y x '".$repo->file."'";
if ($verbose) {
	print STDERR "  system('$zip_cmd')\n";
	print STDERR "  ------------------------\n";
	system($zip_cmd." | sed 's/^/    /'"); #sed for nicer output
	print STDERR "  ------------------------\n";
} else {
	system($zip_cmd." >/dev/null 2>&1");
}

# some handy shortcuts to files
my $file_mod_xml    = "$tmp_dir_exp/$repo_name/Mods.xml";
my $file_addons_xml = "$tmp_dir_exp/$repo_name/Addons.xml";
my $file_server_xml = "$tmp_dir_exp/$repo_name/Server.xml";

# test the files for existence
for my $f ($file_mod_xml, $file_addons_xml, $file_server_xml) {
	if (! -r $f) {
		print STDERR "  error: $f not readable!\n";
		cleanup($tmp_dir_exp);
		exit(1);
	}
}


if ($verbose) { print STDERR "  chdir to $target_dir\n";}
chdir($target_dir);


#
# IF LIST MODE: Just give infos and exit
#
if ($list_mode) {
	# generic info
	print "RepoName: $repo_name\n";
	my $xml_parser = new XML::Simple (KeyAttr=>[]);
        my $data = $xml_parser->XMLin($file_server_xml);
	my @print_data = ("Name", "HomePage", "Game", "GameIP", "GamePort", "GamePwd", "RequiredAddons", "OptionalAddons", "MessageOfTheDay",
			"TSIp", "TSChannel", "TSChannelPwd", "TSNickname", "TSAnonimous", "TSPort", "TSUser", "TSPwd", "TSVersion"); 

	foreach my $item (@print_data) {
		if (!ref($data->{Server}->{$item})) {
			print "$item: ".$data->{Server}->{$item}."\n";
		}
        }

	# list mods
	my $xml_parser = new XML::Simple (KeyAttr=>[]);
	my $data = $xml_parser->XMLin($file_mod_xml);
	foreach my $mod (@{$data->{Mods}}) {
		print "Mod: ".$mod->{Name}."\n";
	}

	cleanup($tmp_dir_exp);
	exit(0);
}



#
# NORMAL MODE:
# loop all mods, if glob pattern matches, process
#
my $xml_parser_mods = new XML::Simple (KeyAttr=>[]);
my $data_mod = $xml_parser_mods->XMLin($file_mod_xml);
my $xml_parser_addons = new XML::Simple (KeyAttr=>[]);
my $data_addons = $xml_parser_addons->XMLin($file_addons_xml);
my %glob_matches; # to record, which glob pattern did match how many mods, so we can detect if a selected mod isn't serverd
foreach my $mod (@{$data_mod->{Mods}}) {
	foreach my $mod_glob (@selected_mods) {
		if (!exists($glob_matches{$mod_glob})) {$glob_matches{$mod_glob} = 0; }
		my $pat = glob2pat($mod_glob);
		if ($mod->{Name} =~ /^$pat$/) {
			print STDERR "processing ".$mod->{Name}."\n";
  			$glob_matches{$mod_glob} = 1;

			
			# check files in Addons.xml. I know this introduces obsolete loops, but we
			# process each mods files sequential this way, which enables nicer output.
			my %modfiles = []; # empty hash to hold all files, so we may look them up more easily and faster
			foreach my $addon_file (@{$data_addons->{Addons}}) {
				if ($addon_file->{Path} =~ /^$pat$|^$pat\\/) {
					# Get file state
					# The remote URL is either a fully qualified URL or just relative, so we need to constructan URL it then
					my $remote_md5 = $addon_file->{Md5};
					my $remote_url = $addon_file->{Url} =~ m#^[a-z]+://# ? $addon_file->{Url} : $repo->scheme."://".$repo->host.$repo->path.$addon_file->{Url}; 
					my $local_path = $addon_file->{Path}."/".$addon_file->{Pbo};
					my $local_dir  = $addon_file->{Path};
					$local_dir  =~ s/\\/\//g; # swap backslashes to linux
					$local_path =~ s/\\/\//g; # swap backslashes to linux

					my $local_state = "UNKNOWN";
					my $local_md5   = "none";

					$modfiles{$local_path} = 1; #record file as "modfile"
					
					if (! -e $local_path && ! -f $local_path) {
						# file does not exist locally, state = NEW
						$local_state = "NEW";
					} else {
                                                # file is already here, check MD5
                                		if (!open(MD5_FILE, "<$local_path")) {
							print STDERR "  ERROR: [MD5] Can't open '$local_path': $!";
							$local_state = "NEW"; # error fallback: get the file as new
						} else {
			                                binmode(MD5_FILE);
        	        		                $local_md5 = uc(Digest::MD5->new->addfile(*MD5_FILE)->hexdigest);
						}
                	                	close(MD5_FILE);
						
						if ($local_md5 ne $remote_md5) {
							$local_state = "CHANGED";
						} else {
							$local_state = "OK";
						}
                                        }
					printf "%-65s %-7s (MD5: remote=%s  local=%s)\n", $local_path, $local_state, $remote_md5, $local_md5;

					# if needed, get and unzip the file
					if ($local_state =~ /NEW|CHANGED/ && !$check_mode) {
						if ($verbose) {print "  fetching file... ";}
						my $repo_filecontent = File::Fetch->new(uri => $remote_url);
						my $zipfile          = $repo_filecontent->fetch( to => "$local_dir" );
						if (!$zipfile) {
						        print "  download error!\n";
					        	next;
						} else {
							chdir($local_dir);
							if (-e $local_path) {
								# remove existing file, so we can "overwrite", because 7zr does not allow that
								if ($verbose) {print "unlink(".$local_path.")  ";}
								unlink($local_path);
							}
							if ($verbose) {print "uncompress... ($zipfile)  ";}
							my $zip_cmd = "7zr -y x '".$zipfile."'";
							if ($verbose) {
							        print STDERR "\n    system('$zip_cmd')\n";
							        print STDERR "    ------------------------\n";
							        system($zip_cmd." | sed 's/^/      /'"); #sed for nicer output
							        print STDERR "    ------------------------\n";
							} else {
							        system($zip_cmd." >/dev/null 2>&1");
							}
							unlink($zipfile);
							if ($verbose) { print "  unlink(".$zipfile.")\n"; }

						}
						chdir($target_dir);
						if ($verbose){print " OK\n";}
					}


				} else {
					# for debugging purposes: Show files that did not match the file regex above
					# produces heavy output under some circumstances.
					if ($debug) { print "  pattern '$pat' did not match file entry $addon_file->{Path}.\n"; }
				}

        		}

			# if copy keyfiles is enabled, do it now
			if ($key_copy) {
				chdir($target_dir);
				if (! -d $keyfolder ) {
					if ($verbose) {	print STDERR "created not existing keyfile folder '$keyfolder'\n"; }
					mkpath($keyfolder);
				}
				find ( sub { #thats the "wanted" callback, inline
                                	next unless -f;
					my $_file = $_;
					my $_path = $File::Find::dir;
					$_path =~ s/^\.\///; #strip leading "./"
					my $local_path = "$_path/$_file";
					next unless $local_path =~ /\.bisign/; #process only .bisign
					print "copy keyfile $_file to $keyfolder...";
					if (! copy($_file, "$keyfolder/$_file") ) {
						print "ERROR: $!\n";
					} else {
						print "OK\n";
					}
				}, $mod->{Name} );
			}


			# look for orphaned files in this mod
			chdir($target_dir);
			
			find ( sub { #thats the "wanted" callback, inline
                                next unless -f; # TODO: this will leave directorys blank...
                                my $_file = $_;
                                my $_path = $File::Find::dir;
                                $_path =~ s/^\.\///; #strip leading "./"
				my $local_path = "$_path/$_file";


				next if exists $modfiles{$local_path}; # only process files not processed yet

				my $local_state = "ORPHAN";
				my $remote_md5  = "none";
				my $local_md5   = "";
				if (!open(MD5_FILE, "<$_file")) {
					if ($verbose) { print STDERR "  ERROR: [MD5] $!\n"; }
					$local_md5 = "error";
				} else {
					binmode(MD5_FILE);
					$local_md5 = uc(Digest::MD5->new->addfile(*MD5_FILE)->hexdigest);
				}
				close(MD5_FILE);

				printf "%-65s %-7s (MD5: remote=%s  local=%s)\n", $local_path, $local_state, $remote_md5, $local_md5;

				if ($del_orphan) {
					if ($verbose) { print STDERR "  unlink($_file)\n"; }
					unlink($_file)
				}

                        }, $mod->{Name} );
		}
	}
}

# print put unmatched globs
foreach my $key (keys(%glob_matches)) {
	if ($glob_matches{$key} == 0) {
		print STDERR "warning: ".$key." did not match any of the repository mods!\n";
	}
}





# cleanup
cleanup($tmp_dir_exp);


# done!
#------------------------------------------------------------

# cleanup tmp data
sub cleanup {
	my $path = shift;
	if ($verbose) {	print STDERR "cleanup $tmp_dir_exp..."; }
	rmtree("$tmp_dir_exp");
	if ($verbose) { print STDERR " OK\n"; }
}


# Usage information for help screen
sub usage {
        print "Sync with a Yoma Addonsync repository\n";
        print "Synopsis:\n";
	print "  ./yas_makeRepo.pl [-S|-L|-C] [options]\n";
        print "  ./yas_makeRepo.pl -h\n";
        print "\nMandatory options:\n";
	print "  -u  URL to repository 7z (AutoConf URL)\n";
        print "\nOptional options:\n";
	print "  -a  Path where the mods should be placed (ie. arma install dir). Default is '$target_dir'\n";
	print "      If given relative, the base is considered the place where yas_makeRepo.pl lives.\n";
	print "  -d  Delete orphaned files\n";
	print "  -k  Copy mods keyfiles (*.bisign) to \"\$target_dir/keys/\"\n";
	print "  -m  Mods to sync. Default is to sync all mods available in the repository.\n";
	print "      Specify like arma modline, wildcard supported (-m \"\@cba;\@ace*\")\n";
	print "  -v  Be more verbose (list processed files etc)\n";
	print "  -V  Debug output\n";
        print "  -h  Show help\n";
	print "\nMode selection:\n";
	print "  -S  default: sync repository\n";
	print "  -C  Check mode; same as normal operation but do not download or delete files\n";
	print "  -L  List repository information (available mods etc) and exit\n";
}

# Prints extendet help
sub help {
	usage();
}

# sub to convert a shell glob pattern into a regular expression
# taken from http://docstore.mik.ua/orelly/perl/cookbook/ch06_10.htm
# use like $pat = glob2pat("foo*");
sub glob2pat {
	my $globstr = shift;
	my %patmap  = (
		'*' => '.*',
		'?' => '.',
		'[' => '[',
		']' => ']',
	);

	# test each character and if match found, replace it with the string defined for it in %patmap
	$globstr =~ s{(.)} { $patmap{$1} || "\Q$1" }ge;

	return $globstr;
}
